; The BANKBOOT provides 2x 8 KB ROM from 00000 to 07fff that
; hide regular RAM in that address space.  RAM starting at 08000
; is available.  To load images from BANKBOOT to RAM at 00000 a
; small loader and the code must first be moved to 08000 or above,
; then BANKBOOT is disabled and the code only is moved again to
; 00000.
;
; Due to the SC/MP 12-bit pointer increment, everything is copied
; in 256 byte chunks aligned to chunk borders.  The copy loop is
; unrolled to copy two bytes per iteration, because it is still
; slow enough that way.
;
; Append up to codelen bytes to bootrom.bin (e.g. KITBUG or NIBL
; ROM image) and flash the result to a boot ROM.  The appended
; code will be moved to RAM at address 0 and started.

  intsyntax -0oct,+0hex,+x'hex'
  bigendian on

l function n,n&0ff
h function n,n>>8

  org 0

  nop

dst1 = 08000
dst2 = 0
codelen = 02000
bankboot = 0c8       ; port address of BANKBOOT
portio = 0ff         ; MSB for IO

; First stage: Move everything from low ROM to high RAM.  Only
; copying the second stage loader and the code would not
; maintain the 256 byte alignment.

  ldi l(0)
  xpal p1
  ldi h(0)
  xpah p1
  ldi 0
  xpal p2
  ldi h(dst1)
  xpah p2
  ldi (256+codelen)/256
  xpal p3
  ldi 0
  xpah p3

copyloop1:
  ccl
  ldi 2
copypage1:
  xae
  ld @1(p1)     ; copy from source to destination
  st @1(p2)
  ld @1(p1)     ; copy from source to destination
  st @1(p2)
  lde
  adi 2
  jnz copypage1
		; The low bytes of p1 and p2 are 254 now
  ld 0(p1)	; copy next to last byte of page
  st 0(p2)
  ld 1(p1)	; copy last byte of page
  st 1(p2)
 
  ccl           ; 16 bit addition of 2 to p1
  xpal p1
  adi 2
  xpal p1
  xpah p1
  adi 0
  xpah p1

  ccl           ; 16 bit addition of 2 to p2
  xpal p2
  adi 2
  xpal p2
  xpah p2
  adi 0
  xpah p2

  ccl		; decrement p3
  xpal p3
  adi -1
  xae
  lde
  xpal p3
  lde
  jnz copyloop1

; Jump to the second stage at dst1.  It works like this, because
; the first stage is copied as well and second_stage does not
; start at the beginning of a new 4k page.

  ldi l(dst1+second_stage-1)
  xpal p3
  ldi h(dst1+second_stage-1)
  xpah p3
  xppc p3

; Second stage: This code is assembled for ROM, but run from RAM.

second_stage:

; Disable BANKBOOT

  ldi portio
  xpah p1
  ldi bankboot
  xpal p1
  ldi 080
  st (p1)

; Copy the code to dst2, which is now regular RAM

  ldi l(dst1+256)
  xpal p1
  ldi h(dst1+256)
  xpah p1
  ldi l(dst2)
  xpal p2
  ldi h(dst2)
  xpah p2
  ldi codelen/256
  xpal p3
  ldi 0
  xpah p3

copyloop2:
  ccl
  ldi 2
copypage2:
  xae
  ld @1(p1)     ; copy from source to destination
  st @1(p2)
  ld @1(p1)     ; copy from source to destination
  st @1(p2)
  lde
  adi 2
  jnz copypage2
		; The low bytes of p1 and p2 are 254 now
  ld 0(p1)	; copy next to last byte of page
  st 0(p2)
  ld 1(p1)	; copy last byte of page
  st 1(p2)
 
  ccl           ; 16 bit addition of 2 to p1
  xpal p1
  adi 2
  xpal p1
  xpah p1
  adi 0
  xpah p1

  ccl           ; 16 bit addition of 2 to p2
  xpal p2
  adi 2
  xpal p2
  xpah p2
  adi 0
  xpah p2

  ccl		; decrement p3
  xpal p3
  adi -1
  xae
  lde
  xpal p3
  lde
  jnz copyloop2

; Jump to dst2+1

  ldi l(dst2)
  xpal p3
  ldi h(dst2)
  xpah p3
  xppc p3

; Align to end of 256 byte page

  org 0ff
  db 0
